package de.unikoblenz.fgbks.base.utils.boundary.bicreater;

import de.unikoblenz.fgbks.base.utils.boundary.impl.StringBoundary;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

/**
 * The combine creator implementation of {@link BoundaryBiCreaterType#COMBINE} for {@link
 * StringBoundary}. <br> The result is a new boundary with the combined range of both boundaries, or
 * an empty optional, if the creation is not possible. <br> In: {@code b1: "a"; b2: "b"} <br> Out:
 * {@code "a","b"} <br>
 */
public class BiCreaterStringCombine extends AbstractBoundaryBiCreater<StringBoundary> {

  private static final BiCreaterStringCombine instance = new BiCreaterStringCombine();

  private BiCreaterStringCombine() {
    super();
  }

  /**
   * Get the current instance of this class.
   *
   * @return the singleton instance
   */
  public static BiCreaterStringCombine getInstance() {
    return instance;
  }

  @Override
  public BoundaryBiCreaterType getType() {
    return BoundaryBiCreaterType.COMBINE;
  }

  @Override
  public Optional<StringBoundary> create(StringBoundary b1, StringBoundary b2) {
    if (b1.matchesAny() || b2.matchesAny()) {
      return Optional.of(StringBoundary.getBuilder().matchesAny(true).build());
    }
    if (!b1.matchesNoneOfValues() && !b2.matchesNoneOfValues()) {
      return Optional.of(
          StringBoundary.getBuilder().addValues(b1.getValues()).addValues(b2.getValues()).build());
    }
    Set<String> values = new HashSet<>();
    if (b1.matchesNoneOfValues() && b2.matchesNoneOfValues()) {
      values = intersection(b1.getValues(), b2.getValues());
    } else if (b1.matchesNoneOfValues()) {
      values = leftInter(b1.getValues(), b2.getValues());
    } else if (b2.matchesNoneOfValues()) {
      values = leftInter(b2.getValues(), b1.getValues());
    }
    return Optional.of(
        StringBoundary.getBuilder()
            .matchesNoneOfValues(true)
            .addValues(values)
            .matchesAny(values.size() == 0)
            .build());
  }

  private Set<String> leftInter(String[] values1, String[] values2) {
    Set<String> vals = new HashSet<>(Arrays.asList(values1));
    for (String v : values2) {
      vals.removeIf(x -> x.equals(v));
    }
    return vals;
  }

  private Set<String> intersection(String[] values1, String[] values2) {
    Set<String> vals = new HashSet<>();
    for (String v1 : values1) {
      for (String v2 : values2) {
        if (v1.equals(v2)) {
          vals.add(v1);
        }
      }
    }
    return vals;
  }
}
