package de.unikoblenz.fgbks.base.utils.boundary.checker;

import static de.unikoblenz.fgbks.base.utils.boundary.BoundType.INCLUSIVE;

import de.unikoblenz.fgbks.base.utils.boundary.AbstractGrowingBoundary;

/**
 * The in contact checker implementation of {@link BoundaryCheckType#IS_IN_CONTACT} for {@link
 * AbstractGrowingBoundary}. <br> The check returns true, if the first boundary is in contact with
 * the other boundary.
 * <br>
 * E.g: <br>
 *
 * <ul>
 *   <li>b1 = "= 1"; b2 = "= 1" -{@literal >} true
 *   <li>b1 = "{@literal <}= 1"; b2 = "{@literal <} 2" -{@literal >} true
 *   <li>b1 = "[0..20]"; b2 = "]20..22]" -{@literal >} false
 * </ul>
 */
public class CheckInContact extends AbstractBoundaryCheck<AbstractGrowingBoundary> {

  private static final CheckInContact instance = new CheckInContact();

  private CheckInContact() {
    super();
  }

  /**
   * Get the current instance of this class.
   *
   * @return the singleton instance
   */
  public static CheckInContact getInstance() {
    return instance;
  }

  @Override
  public BoundaryCheckType getType() {
    return BoundaryCheckType.IS_IN_CONTACT;
  }

  @Override
  public boolean check(AbstractGrowingBoundary b1, AbstractGrowingBoundary b2) {
    return checkInContact(b1, b2);
  }

  /**
   * see {@link CheckInContact}
   */
  public static boolean checkInContact(AbstractGrowingBoundary b1, AbstractGrowingBoundary b2) {
    // TODO: check raw types
    if (b1.getLowerBound().compareTo(b2.getLowerBound()) <= 0) {
      return checkOrdered(b1, b2);
    } else {
      return checkOrdered(b2, b1);
    }
  }

  private static boolean checkOrdered(AbstractGrowingBoundary b1, AbstractGrowingBoundary b2) {
    if (b1.getUpperBound().compareTo(b2.getLowerBound()) > 0) {
      return true;
    }
    if (b1.getUpperBound().compareTo(b2.getLowerBound()) == 0) {
      return b1.getUpperBoundType() == INCLUSIVE && b2.getLowerBoundType() == INCLUSIVE;
    }
    return false;
  }
}
