package de.unikoblenz.fgbks.core.dmn.domain.vdmn;

import de.unikoblenz.fgbks.base.domain.Label;
import de.unikoblenz.fgbks.base.domain.Name;
import de.unikoblenz.fgbks.base.domain.StringValue;
import de.unikoblenz.fgbks.core.dmn.domain.ids.ColumnId;
import java.util.List;
import java.util.Optional;

/**
 * A column is a abstract interface for the {@link VDmnInputColumn} and {@link VDmnOutputColumn}. A
 * column defines the {@link VTypeRef} of the containing values.
 */
public interface VDmnColumn extends VDmnElement {

  /**
   * Get the {@link VDmnDecisionTable}.
   *
   * @return the {@link VDmnDecisionTable}
   */
  VDmnDecisionTable getDmnDecisionTable();

  /**
   * Check, if the given column is a {@link VDmnInputColumn}. If this method returns true, than
   * {@link VDmnColumn#isOutputColumn()} returns false.
   *
   * @return true, if the current column is a input column
   */
  default boolean isInputColumn() {
    return this instanceof VDmnInputColumn;
  }

  /**
   * Check, if the given column is a {@link VDmnOutputColumn}. If this method returns true, than
   * {@link VDmnColumn#isInputColumn()} returns false.
   *
   * @return true, if the current column is a output column
   */
  default boolean isOutputColumn() {
    return this instanceof VDmnOutputColumn;
  }

  /**
   * Get the list of predefined {@link StringValue}s, if the column has the type {@link
   * VTypeRef#STRING} and the model contains these values.
   *
   * @return the list of {@link StringValue}s or an empty optional
   */
  Optional<List<StringValue>> getPredefinedValues();

  /**
   * Get the {@link Name}, if present.
   *
   * @return the {@link Name} or an empty optional
   */
  Optional<Name> getName();

  /**
   * Get the {@link Label}, if present.
   *
   * @return the {@link Label} or an empty optional
   */
  Optional<Label> getLabel();

  /**
   * Get the {@link VTypeRef}.
   *
   * @return the {@link VTypeRef}
   */
  VTypeRef getTypeRef();

  /**
   * Get the list of {@link VDmnValue}s.
   *
   * @return the list of {@link VDmnValue}s
   */
  List<VDmnValue> getValues();

  /**
   * Get the {@link ColumnId}.
   *
   * @return the {@link ColumnId}
   */
  default ColumnId getColumnId() {
    return (ColumnId) getId();
  }

  /**
   * Get the {@link VDmnDecision}.
   *
   * @return the {@link VDmnDecision}
   */
  default VDmnDecision getDmnDecision() {
    return getDmnDecisionTable().getDmnDecision();
  }

  /**
   * Get the {@link VDmnDefinition}.
   *
   * @return the {@link VDmnDefinition}
   */
  default VDmnDefinition getDmnDefinition() {
    return getDmnDecisionTable().getDmnDefinition();
  }
}
