package de.unikoblenz.fgbks.core.dmn.domain.vdmn.impl;

import de.unikoblenz.fgbks.core.dmn.domain.ids.AbstractId;
import de.unikoblenz.fgbks.core.dmn.domain.ids.DefinitionId;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDecision;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDecisionTable;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDefinition;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnElement;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnInputData;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.apache.commons.lang3.Validate;

/**
 * Implementation of {@link VDmnDefinition}.
 */
public class VDmnDefinitionImpl extends AbstractVDmnElement implements VDmnDefinition {

  private List<VDmnDecision> dmnDecisions;
  private List<VDmnInputData> dmnInputData;
  private Map<AbstractId, VDmnElement> elements;

  private VDmnDefinitionImpl() {
    super();
    dmnDecisions = new ArrayList<>();
    dmnInputData = new ArrayList<>();
    elements = new HashMap<>();
  }

  @Override
  public List<VDmnDecision> getDmnDecisions() {
    return Collections.unmodifiableList(dmnDecisions);
  }

  @Override
  public List<VDmnInputData> getDmnInputData() {
    return Collections.unmodifiableList(dmnInputData);
  }

  @Override
  public VDmnElement getElementById(AbstractId id) {
    return elements.get(Validate.notNull(id));
  }

  /**
   * Get a new instance of a builder.
   *
   * @return a new builder instance
   */
  public static Builder getBuilder() {
    return new VDmnDefinitionImpl().new Builder();
  }

  public class Builder extends AbstractDmnElementBuilder<VDmnDefinitionImpl> {

    public Builder withDefinitionId(DefinitionId definitionId) {
      value.id = definitionId;
      return this;
    }

    public Builder addDmnDecision(VDmnDecision dmnDecision) {
      value.dmnDecisions.add(Validate.notNull(dmnDecision));
      return this;
    }

    public Builder addDmnDecisionFromBuilder(VDmnDecisionImpl.Builder dmnDecisionBuilder) {
      addDmnDecision(dmnDecisionBuilder.getUnbuildValue());
      dmnDecisionBuilder.withDmnDefinition(value);
      return this;
    }

    public Builder addDmnInputData(VDmnInputData dmnInputData) {
      value.dmnInputData.add(Validate.notNull(dmnInputData));
      return this;
    }

    public Builder addDmnInputDataFromBuilder(VDmnInputDataImpl.Builder dmnInputDataBuilder) {
      addDmnInputData(dmnInputDataBuilder.getUnbuildValue());
      dmnInputDataBuilder.withDmnDefinition(value);
      return this;
    }

    @Override
    protected void validate() {
      super.validate();
      Validate.notNull(value.id);
    }

    @Override
    public VDmnDefinitionImpl build() {
      // save all Ids and elements in the element map
      // add the definition
      value.elements.put(getId(), this.value);
      // add the input data
      value.getDmnInputData().forEach(inData -> value.elements.put(inData.getId(), inData));
      // add all decisions
      for (VDmnDecision decision : value.getDmnDecisions()) {
        value.elements.put(decision.getId(), decision);
        // add the decision table
        VDmnDecisionTable table = decision.getDmnDecisionTable();
        value.elements.put(table.getId(), table);
        // add columns
        table.getInputColumns().forEach(column -> value.elements.put(column.getId(), column));
        table
            .getInputColumns()
            .forEach(column -> value.elements.put(column.getInputExpressionId(), column));
        table.getOutputColumns().forEach(column -> value.elements.put(column.getId(), column));
        // add rules
        table.getRules().forEach(rule -> value.elements.put(rule.getId(), rule));
        // add input values/enties
        table.getInputColumns().stream()
            .flatMap(column -> column.getInputValues().stream())
            .forEach(iv -> value.elements.put(iv.getId(), iv));
        table.getOutputColumns().stream()
            .flatMap(column -> column.getDmnOutputValues().stream())
            .forEach(iv -> value.elements.put(iv.getId(), iv));
      }
      return super.build();
    }
  }
}
