package de.unikoblenz.fgbks.core.dmn.verification;

import de.unikoblenz.fgbks.base.utils.UniqueIdGenerator;
import de.unikoblenz.fgbks.core.dmn.utils.DmnObjectContainer;
import de.unikoblenz.fgbks.core.dmn.utils.DmnService;
import de.unikoblenz.fgbks.core.dmn.verification.metrics.DmnVerificationMetricsService;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerifierResult;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerifierResultSet;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerifierResultSet.Builder;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.AbstractVerifier;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.DmnVerifier;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.Verifier;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.classification.Classification;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.classification.ClassificationType;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.config.AbstractDmnVerifierConfig;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.config.DefaultConfiguration;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.types.Type;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.types.VerificationType;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import javax.enterprise.context.ApplicationScoped;
import javax.inject.Inject;
import org.apache.commons.lang3.Validate;
import org.eclipse.microprofile.config.inject.ConfigProperty;
import org.reflections.Reflections;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Service class for generating verifications of DMNs. This class scans for verifier, witch are
 * annotated with {@link DmnVerifier}. The verifier can be executed by calling {@link
 * DmnVerificationService#generate} or by filtering the {@link VerificationType} (and annotated with
 * {@link Type}) or {@link ClassificationType} (and annotated with {@link Classification}).
 */
@ApplicationScoped
public class DmnVerificationService {

  private static final Logger LOGGER =
      LoggerFactory.getLogger(DmnVerificationService.class.getSimpleName());

  @ConfigProperty(name = "verifier.threads", defaultValue = "2")
  Integer maxThreads;

  @ConfigProperty(
      name = "verifier.package",
      defaultValue = "de.unikoblenz.fgbks.core.dmn.verification.verifier")
  String packagePath;

  private ExecutorService executor;
  private Set<Class<?>> verifierClasses;
  private List<VerificationType> verificationTypes;
  private List<ClassificationType> verificationClassificationTypes;

  @Inject protected DmnService dmnService;
  @Inject
  protected DmnVerificationMetricsService dmnVerificationMetricsService;

  protected DmnVerificationService() {
    super();
  }

  /**
   * Create a {@link VerifierResultSet} by executing all verifiers, which are registered in the
   * current deployment.
   *
   * @param dmnXml the XML of the DMN as String
   * @return a {@link VerifierResultSet} containing all verifications of known verifiers.
   */
  public VerifierResultSet generate(String dmnXml) {
    return generateWithFilter(dmnXml, null, null);
  }

  /**
   * Create a {@link VerifierResultSet} by executing the verifiers, which is registered in the
   * current deployment and has the name of the {@code type} parameter.
   *
   * @param type the type as String. Should be the name of a {@link VerificationType}.
   * @param dmnXml the XML of the DMN as String
   * @return a {@link VerifierResultSet} containing all verifications of the selected verifier.
   */
  public VerifierResultSet generateFromType(String type, String dmnXml) {
    return generateWithFilter(dmnXml, type, null);
  }

  /**
   * Create a {@link VerifierResultSet} by executing the verifiers, which is registered in the
   * current deployment and has the name of the {@code type} parameter.
   *
   * @param classification the type as String. Should be the name of a {@link ClassificationType}.
   * @param dmnXml the XML of the DMN as String
   * @return a {@link VerifierResultSet} containing all verifications of the selected verifier.
   */
  public VerifierResultSet generateFromClassification(String classification, String dmnXml) {
    return generateWithFilter(dmnXml, null, classification);
  }

  /**
   * This function returns all currently registered {@link VerificationType}s as {@link
   * Collections#unmodifiableList}.
   *
   * @return a list of {@link VerificationType}s
   */
  public List<VerificationType> getVerificationTypes() {
    return Collections.unmodifiableList(verificationTypes);
  }

  /**
   * This function returns all currently registered {@link ClassificationType}s as {@link
   * Collections#unmodifiableList}.
   *
   * @return a list of {@link ClassificationType}s
   */
  public List<ClassificationType> getVerificationClassificationTypes() {
    return Collections.unmodifiableList(verificationClassificationTypes);
  }

  /**
   * Init the verification service instance.
   */
  public void initVerifier() {
    if (executor == null || verifierClasses == null) {
      LOGGER.info("Init verification.");
      LOGGER.info("Creating thread pool with " + maxThreads + " threads.");
      executor = Executors.newFixedThreadPool(maxThreads);

      LOGGER.info("Init verifier.");
      Reflections reflections = new Reflections(packagePath);
      verifierClasses = reflections.getTypesAnnotatedWith(DmnVerifier.class);

      verificationTypes = new ArrayList<>();
      Set<Class<?>> clVts = reflections.getTypesAnnotatedWith(Type.class);
      for (Class<?> clVt : clVts) {
        try {
          VerificationType verificationType =
              (VerificationType) clVt.getMethod("getInstance").invoke(null, null);
          LOGGER.info(
              "Verifier found: "
                  + verificationType.getName()
                  + " ("
                  + verificationType.getClassification().getName()
                  + ")");
          verificationTypes.add(verificationType);
        } catch (IllegalAccessException | NoSuchMethodException | InvocationTargetException e) {
          e.printStackTrace();
        }
      }

      verificationClassificationTypes = new ArrayList<>();
      Set<Class<?>> clCts = reflections.getTypesAnnotatedWith(Classification.class);
      for (Class<?> clCt : clCts) {
        try {
          verificationClassificationTypes.add(
              (ClassificationType) clCt.getMethod("getInstance").invoke(null, null));
        } catch (IllegalAccessException | NoSuchMethodException | InvocationTargetException e) {
          e.printStackTrace();
        }
      }
    }
  }

  private Map<Class<? extends AbstractDmnVerifierConfig>, AbstractDmnVerifierConfig> getConfigs(
      String dmnXml, long verificationProcessId) {
    Map<Class<? extends AbstractDmnVerifierConfig>, AbstractDmnVerifierConfig> map =
        new ConcurrentHashMap<>();
    Optional<DmnObjectContainer> dmnObjectContainerFromXml =
        dmnService.getDmnObjectContainerFromXml(dmnXml);
    if (!dmnObjectContainerFromXml.isPresent()) {
      throw new IllegalArgumentException("Dmn table can not be parsed.");
    } else {
      map.put(
          DefaultConfiguration.class,
          DefaultConfiguration.getBuilder()
              .withVerificationProcessId(verificationProcessId)
              .withDmnObjectContainer(dmnObjectContainerFromXml.get())
              .build());
      return map;
    }
  }

  private VerifierResultSet generateWithFilter(
      String dmnXml, String typeName, String classificationName) {
    long verificationProcessId = UniqueIdGenerator.getNextId();
    LOGGER.info("Start verification process. Id: " + verificationProcessId);
    LOGGER.info("Parsing dmn. Id: " + verificationProcessId);
    Validate.notNull(dmnXml);
    LOGGER.debug(dmnXml);
    Builder resultBuilder = VerifierResultSet.getBuilder();
    Map<Class<? extends AbstractDmnVerifierConfig>, AbstractDmnVerifierConfig> configs;
    try {
      configs = getConfigs(dmnXml, verificationProcessId);
    } catch (IllegalArgumentException e) {
      LOGGER.warn("Parsing dmn failed. Id: " + verificationProcessId);
      return resultBuilder.build();
    }
    LOGGER.info("Looking for verifiers. Id: " + verificationProcessId);
    List<Future<VerifierResult>> results = new ArrayList<>(verifierClasses.size());

    List<Verifier> executedVerifier = new ArrayList<>();
    for (Class<?> verifierClass : verifierClasses) {
      AbstractVerifier av = null;
      try {
        av = (AbstractVerifier) verifierClass.newInstance();
      } catch (InstantiationException | IllegalAccessException e) {
        LOGGER.error(e.getMessage());
      }
      if (av != null
          && (typeName == null || av.getVerificationType().getName().getValue().equals(typeName))
          && (classificationName == null
              || av.getVerificationType()
                  .getClassification()
                  .getName()
                  .getValue()
                  .equals(classificationName))) {
        // add verifier to list
        executedVerifier.add(av);
        // Create Configuration
        AbstractDmnVerifierConfig verifierConfig =
            configs.get(av.getClass().getAnnotation(DmnVerifier.class).verifierConfig());

        // Add Configuration and start Calculation
        results.add(av.withConfiguration(verifierConfig).verify(executor));
      }
    }
    // add all verifier results to the result Builder
    results.forEach(
        r -> {
          try {
            resultBuilder.addVerifierResult(r.get());
          } catch (InterruptedException | ExecutionException e) {
            e.printStackTrace();
          }
        });
    LOGGER.info("Verification finished. Id: " + verificationProcessId);

    // add execution times to metric statistics
    executedVerifier.forEach(
        v ->
            dmnVerificationMetricsService.addExecutionTime(
                v.getVerificationType(), v.getExecutionTime()));

    return resultBuilder.build();
  }
}
