package de.unikoblenz.fgbks.core.dmn.verification.result;

import de.unikoblenz.fgbks.base.builder.DefaultBuilder;
import de.unikoblenz.fgbks.base.domain.Message;
import de.unikoblenz.fgbks.core.dmn.verification.result.actions.VerificationFix;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.Verifier;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import javax.json.bind.annotation.JsonbProperty;
import org.apache.commons.lang3.Validate;

/**
 * A verification result entry represents on single error, or warning, which was calculated by one
 * {@link Verifier}. One {@link VerificationResultEntry} contains a set of {@link
 * VerificationResultEntryElement}s. These elements represents one rule / one table / one column or
 * one cell of the dmn.
 */
public class VerificationResultEntry extends AbstractResultObject {

  private List<VerificationFix> verificationFixes;
  private Set<VerificationResultEntryElement> verificationResultEntryElements;
  private Message message;
  private VerificationClassification verificationClassification;

  protected VerificationResultEntry() {
    this.verificationResultEntryElements = ConcurrentHashMap.newKeySet();
    verificationFixes = new ArrayList<>();
  }

  /**
   * Get the list of {@link VerificationFix}es.
   *
   * @return the list of {@link VerificationFix}es
   */
  @JsonbProperty("verificationFixes")
  public List<VerificationFix> getVerificationsFixes() {
    return Collections.unmodifiableList(verificationFixes);
  }

  /**
   * Get the amount of verification result entry elements.
   *
   * @return the amount of verification result entry elements
   */
  @JsonbProperty("size")
  private int getAmountOfElements() {
    return verificationResultEntryElements.size();
  }

  /**
   * Get the {@link VerificationClassification}.
   *
   * @return the {@link VerificationClassification}
   */
  @JsonbProperty("verificationClassification")
  public VerificationClassification getVerificationClassification() {
    return verificationClassification;
  }

  /**
   * Get the set of {@link VerificationResultEntryElement}s.
   *
   * @return the {@link VerificationResultEntryElement}s
   */
  @JsonbProperty("elements")
  public Set<VerificationResultEntryElement> getVerificationResultEntryElements() {
    return new HashSet<>(verificationResultEntryElements);
  }

  /**
   * Enum to classify one {@link VerificationResultEntry}.
   */
  public enum VerificationClassification {
    /** Only information. */
    INFO,
    /** Warning. No action needed. */
    WARNING,
    /** Error. Action needed. */
    ERROR,
    /** Fatal error. Must be fixed, so that the dmn is executable. */
    FATAL_ERROR
  }

  /**
   * Get the message of the verification, which describes the verification textual.
   *
   * @return the {@link Message}
   */
  @JsonbProperty("message")
  public Message getMessage() {
    return new Message(message);
  }

  /**
   * Get a new instance of a {@link Builder} for a {@link VerificationResultEntry}.
   *
   * @return a new {@link Builder}
   */
  public static Builder getBuilder() {
    return new VerificationResultEntry().new Builder();
  }

  /** A builder class for {@link VerificationResultEntry}. */
  public class Builder extends DefaultBuilder<VerificationResultEntry> {

    /**
     * Add a new {@link VerificationResultEntryElement} to the {@link VerificationResultEntry}.
     *
     * @param verificationResultEntryElement the {@link VerificationResultEntryElement}
     * @return the current builder
     */
    public Builder addVerificationResultEntryElement(
        VerificationResultEntryElement verificationResultEntryElement) {
      value.verificationResultEntryElements.add(Validate.notNull(verificationResultEntryElement));
      return this;
    }

    /**
     * Set the {@link Message} of the {@link VerificationResultEntry}.
     *
     * @param message the {@link Message}
     * @return the current builder
     */
    public Builder withMessage(Message message) {
      value.message = new Message(message);
      return this;
    }

    /**
     * Set the {@link VerificationClassification} of the {@link VerificationResultEntry}.
     *
     * @param verificationClassification the {@link VerificationClassification}
     * @return the current builder
     */
    public Builder withClassification(VerificationClassification verificationClassification) {
      value.verificationClassification = verificationClassification;
      return this;
    }

    /**
     * Add a {@link VerificationFix} to the {@link VerificationResultEntry}.
     *
     * @param verificationFix the {@link VerificationFix}
     * @return the current builder
     */
    public Builder addVerificationFix(VerificationFix verificationFix) {
      value.verificationFixes.add(Validate.notNull(verificationFix));
      return this;
    }

    @Override
    protected void validate() {
      super.validate();
      Validate.notNull(value.message);
      Validate.notNull(value.verificationClassification);
      Validate.notEmpty(value.verificationResultEntryElements);
    }
  }
}
