package de.unikoblenz.fgbks.core.dmn.verification.result;

import de.unikoblenz.fgbks.core.dmn.domain.ids.AbstractId;
import de.unikoblenz.fgbks.core.dmn.domain.ids.JsonIdentifier;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnColumn;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDecision;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnDecisionTable;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnInputData;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnNode;
import de.unikoblenz.fgbks.core.dmn.domain.vdmn.VDmnValue;
import java.util.HashMap;
import java.util.Map;
import javax.json.bind.annotation.JsonbProperty;
import org.apache.commons.lang3.Validate;

/**
 * A verification result entry element represents one dmn tabel / rule / column or single cell of
 * the dmn. The identification of this elements works with {@link
 * VerificationResultEntryElement#getIdentifier()}.
 */
public class VerificationResultEntryElement extends AbstractResultObject {

  private final Map<String, AbstractId> identifier;

  /**
   * Create a new (empty) {@link VerificationResultEntryElement}. Call {@link
   * VerificationResultEntryElement#withIdentifier(AbstractId)} to add new identifier.
   *
   * @return the new {@link VerificationResultEntryElement}
   */
  public static VerificationResultEntryElement create() {
    return new VerificationResultEntryElement();
  }

  private VerificationResultEntryElement() {
    identifier = new HashMap<>();
  }

  /**
   * Create a new {@link VerificationResultEntryElement} with the initial identifier from a {@link
   * VDmnValue}. Call {@link VerificationResultEntryElement#withIdentifier(AbstractId)} to add
   * further identifier.
   *
   * @param vDmnValue the {@link VDmnValue}
   * @return the new {@link VerificationResultEntryElement}
   */
  public static VerificationResultEntryElement create(VDmnValue vDmnValue) {
    return new VerificationResultEntryElement()
        .withIdentifier(vDmnValue.getDefinitionId())
        .withIdentifier(vDmnValue.getDecisionId())
        .withIdentifier(vDmnValue.getDecisionTableId())
        .withIdentifier(vDmnValue.getDmnColumn().getColumnId())
        .withIdentifier(vDmnValue.getRuleId())
        .withIdentifier(vDmnValue.getId());
  }

  /**
   * Create a new {@link VerificationResultEntryElement} with the initial identifier from a {@link
   * VDmnDecisionTable}. Call {@link VerificationResultEntryElement#withIdentifier(AbstractId)} to
   * add further identifier.
   *
   * @param dmnDecisionTable the {@link VDmnDecisionTable}
   * @return the new {@link VerificationResultEntryElement}
   */
  public static VerificationResultEntryElement create(VDmnDecisionTable dmnDecisionTable) {
    return new VerificationResultEntryElement()
        .withIdentifier(dmnDecisionTable.getDmnDefinition().getDefinitionId())
        .withIdentifier(dmnDecisionTable.getDmnDecision().getDecisionId())
        .withIdentifier(dmnDecisionTable.getDecisionTableId());
  }

  /**
   * Create a new {@link VerificationResultEntryElement} with the initial identifier from a {@link
   * VDmnColumn}. Call {@link VerificationResultEntryElement#withIdentifier(AbstractId)} to add
   * further identifier.
   *
   * @param column the {@link VDmnColumn}
   * @return the new {@link VerificationResultEntryElement}
   */
  public static VerificationResultEntryElement create(VDmnColumn column) {
    return new VerificationResultEntryElement()
        .withIdentifier(column.getDmnDefinition().getDefinitionId())
        .withIdentifier(column.getDmnDecision().getDecisionId())
        .withIdentifier(column.getDmnDecisionTable().getDecisionTableId())
        .withIdentifier(column.getColumnId());
  }

  /**
   * Create a new {@link VerificationResultEntryElement} with the initial identifier from a {@link
   * VDmnInputData}. Call {@link VerificationResultEntryElement#withIdentifier(AbstractId)} to add
   * further identifier.
   *
   * @param inputData the {@link VDmnInputData}
   * @return the new {@link VerificationResultEntryElement}
   */
  public static VerificationResultEntryElement create(VDmnInputData inputData) {
    return new VerificationResultEntryElement()
        .withIdentifier(inputData.getDmnDefinition().getDefinitionId())
        .withIdentifier(inputData.getInputDataId());
  }

  /**
   * Create a new {@link VerificationResultEntryElement} with the initial identifier from a {@link
   * VDmnNode}. Call {@link VerificationResultEntryElement#withIdentifier(AbstractId)} to add
   * further identifier.
   *
   * @param vDmnNode the {@link VDmnInputData} or {@link VDmnDecision} node
   * @return the new {@link VerificationResultEntryElement}
   */
  public static VerificationResultEntryElement create(VDmnNode vDmnNode) {
    return new VerificationResultEntryElement()
        .withIdentifier(vDmnNode.getDmnDefinition().getDefinitionId())
        .withIdentifier(vDmnNode.getId());
  }

  /**
   * Get all required ids ({@link AbstractId}) of the {@link VerificationResultEntryElement}.
   *
   * @return a map with the identifiers
   */
  @JsonbProperty("identifier")
  public Map<String, AbstractId> getIdentifier() {
    return new HashMap<>(identifier);
  }

  /**
   * Add a new {@link AbstractId} to the identifier map. The key for the value (the id) is the name
   * of the concrete class with the annotation {@link JsonIdentifier}. One identifier can not be
   * added twice.
   *
   * @param value the {@link AbstractId}, which must be annotated with the {@link JsonIdentifier}
   *     class
   * @return the current {@link VerificationResultEntryElement}
   */
  public VerificationResultEntryElement withIdentifier(AbstractId value) {

    // Check if annotation is present
    Validate.isTrue(
        value.getClass().isAnnotationPresent(JsonIdentifier.class),
        "No annotation \"JsonIdentifier\" for AbstractId %s present.",
        value.getClass().getSimpleName());
    String keyValue = value.getClass().getAnnotation(JsonIdentifier.class).value();
    // every identifier can only be set once
    Validate.isTrue(
        !identifier.containsKey(keyValue),
        "Identifier %s with value %s in %s already set.",
        keyValue,
        value,
        VerificationResultEntryElement.class.getSimpleName());
    identifier.put(keyValue, value);
    return this;
  }
}
