package de.unikoblenz.fgbks.core.dmn.verification.result;

import de.unikoblenz.fgbks.base.domain.Message;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerificationResultEntry.VerificationClassification;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerifierResult.Builder;
import de.unikoblenz.fgbks.core.dmn.verification.result.actions.VerificationFix;
import org.apache.commons.lang3.Validate;

/**
 * A factory to help creating a {@link VerifierResult}. First, call {@link
 * VerificationResultEntryFactory#create(Builder)} with and instance of the {@link
 * VerifierResult.Builder}. Than, for each new result for one verifier first call {@link
 * VerificationResultEntryFactory#addElement(VerificationResultEntryElement)} for every element you
 * want to add. Finally call {@link
 * VerificationResultEntryFactory#addToEntry(VerificationClassification, String)} to add the result.
 */
public class VerificationResultEntryFactory {

  private VerifierResult.Builder verifierResultBuilder;
  private VerificationResultEntry.Builder verificationResultEntryBuilder;

  private VerificationResultEntryFactory(Builder verifierResultBuilder) {
    this.verifierResultBuilder = Validate.notNull(verifierResultBuilder);
    verificationResultEntryBuilder = null;
  }

  /**
   * Create a new {@link VerificationResultEntryFactory}.
   *
   * @param verifierResultBuilder the instance of a {@link VerifierResult.Builder}
   * @return the new {@link VerificationResultEntryFactory}
   */
  public static VerificationResultEntryFactory create(Builder verifierResultBuilder) {
    return new VerificationResultEntryFactory(verifierResultBuilder);
  }

  private VerificationResultEntry.Builder getCurrentOrCreate() {
    if (verificationResultEntryBuilder == null) {
      verificationResultEntryBuilder = VerificationResultEntry.getBuilder();
    }
    return verificationResultEntryBuilder;
  }

  /**
   * Add a new {@link VerificationFix} to the {@link VerificationResultEntry}.
   *
   * @param verificationFix the {@link VerificationFix}
   * @return the current {@link VerificationResultEntryFactory}
   */
  public VerificationResultEntryFactory addVerificationFix(VerificationFix verificationFix) {
    getCurrentOrCreate().addVerificationFix(verificationFix);
    return this;
  }

  /**
   * Add a {@link VerificationResultEntryElement} to the current {@link VerificationResultEntry}.
   *
   * @param verificationResultEntryElement the {@link VerificationResultEntryElement} to add
   * @return the current {@link VerificationResultEntryFactory}
   */
  public VerificationResultEntryFactory addElement(
      VerificationResultEntryElement verificationResultEntryElement) {
    getCurrentOrCreate().addVerificationResultEntryElement(verificationResultEntryElement);
    return this;
  }

  /**
   * Add the previous added {@link VerificationResultEntryElement}s to the {@link VerifierResult}
   * with the given {@link VerificationClassification} and the given message.
   *
   * @param verificationClassification the {@link VerificationClassification}
   * @param message as String
   */
  public void addToEntry(VerificationClassification verificationClassification, String message) {
    addToEntry(verificationClassification, message, null);
  }

  /**
   * Add the previous added {@link VerificationResultEntryElement}s to the {@link VerifierResult}
   * with the given {@link VerificationClassification} and the given message.
   *
   * @param verificationClassification the {@link VerificationClassification}
   * @param message as String
   * @param args the additional arguments for formatting the message
   */
  public void addToEntry(
      VerificationClassification verificationClassification, String message, Object... args) {
    addToEntry(verificationClassification, new Message(String.format(message, args)));
  }

  /**
   * Add the previous added {@link VerificationResultEntryElement}s to the {@link VerifierResult}
   * with the given {@link VerificationClassification} and the given {@link Message}.
   *
   * @param verificationClassification the {@link VerificationClassification}
   * @param message as {@link Message}
   */
  public void addToEntry(VerificationClassification verificationClassification, Message message) {
    Validate.notNull(
        verificationResultEntryBuilder,
        "First call \"addElement\" in VerificationResultEntryFactory before calling \"addToEntry\"");
    verificationResultEntryBuilder.withClassification(verificationClassification);
    verificationResultEntryBuilder.withMessage(new Message(message));
    verifierResultBuilder.addVerificationResultEntry(verificationResultEntryBuilder.build());
    verificationResultEntryBuilder = null;
  }
}
