package de.unikoblenz.fgbks.core.dmn.verification.verifier;

import de.unikoblenz.fgbks.core.dmn.utils.DmnObjectContainer;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerificationResultEntryFactory;
import de.unikoblenz.fgbks.core.dmn.verification.result.VerifierResult;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.config.AbstractDmnVerifierConfig;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.config.DefaultConfiguration;
import de.unikoblenz.fgbks.core.dmn.verification.verifier.types.VerificationType;
import java.lang.reflect.InvocationTargetException;
import java.util.Objects;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import org.apache.commons.lang3.Validate;
import org.apache.commons.lang3.time.StopWatch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * The abstract verifier is a implementation staring point for the {@link Verifier} interface and
 * contains the needed components for the final implementation of verifiers.
 */
public abstract class AbstractVerifier implements Verifier {

  protected final Logger LOGGER = LoggerFactory.getLogger(getClass().getSimpleName());

  protected VerificationType verificationType;
  protected AbstractDmnVerifierConfig verifierConfig;
  protected VerifierResult.Builder resultBuilder;
  /**
   * The {@link VerificationResultEntryFactory} for adding result entries.
   */
  protected VerificationResultEntryFactory vreFactory;

  protected DmnObjectContainer dmnObjectContainer;

  private long executionTime = -1;

  /** {@inheritDoc} */
  @Override
  public VerificationType getVerificationType() {
    return verificationType;
  }

  /**
   * Load the configuration, which contains the necessary information for the verifier. E.g. the dmn
   * model.
   *
   * @param configuration the configuration object
   * @return the current instance of the verifier
   */
  public final AbstractVerifier withConfiguration(AbstractDmnVerifierConfig configuration) {
    this.verifierConfig = Validate.notNull(configuration);
    this.dmnObjectContainer =
        this.verifierConfig.getConfiguration(DefaultConfiguration.KEY_DMN_OBJECT_CONTAINER);
    return this;
  }

  protected AbstractVerifier() {
    try {
      this.verificationType =
          (VerificationType)
              this.getClass()
                  .getAnnotation(DmnVerifier.class)
                  .verifierType()
                  .getMethod("getInstance")
                  .invoke(null, null);
    } catch (IllegalAccessException | NoSuchMethodException | InvocationTargetException e) {
      e.printStackTrace();
    }
    resultBuilder = VerifierResult.getBuilder().fromVerifier(this);
    vreFactory = VerificationResultEntryFactory.create(resultBuilder);
  }

  /** {@inheritDoc} */
  public final Future<VerifierResult> verify(ExecutorService executor) {
    Validate.notNull(verifierConfig, "Validation config has to been set.");
    Validate.notNull(executor);
    return executor.submit((Callable<VerifierResult>) this::verify);
  }

  private VerifierResult verify() {
    LOGGER.info(
        "Starting verification "
            + getVerificationType().getName()
            + " Id: "
            + this.verifierConfig.getVerificationProcessId());
    StopWatch watch = new StopWatch();
    watch.start();
    doVerification();
    watch.stop();
    executionTime = watch.getNanoTime();
    LOGGER.info(
        "Verification "
            + getVerificationType().getName()
            + " finished. Id: "
            + this.verifierConfig.getVerificationProcessId()
            + " Time Elapsed: "
            + executionTime
            + " nano seconds; Or "
            + executionTime / 1000000
            + " ms");
    return resultBuilder.withExecutionTime(executionTime).build();
  }

  /**
   * This method must be implemented by the finial verifiers and should contain the calculation of
   * the verifier. The {@link AbstractVerifier#vreFactory} can be used for adding result entries.
   */
  protected abstract void doVerification();

  @Override
  public long getExecutionTime() {
    if (executionTime == -1) {
      throw new IllegalStateException(
          "Verifier must be executed before accessing the execution time!");
    }
    return executionTime;
  }

  @Override
  public boolean equals(Object o) {
    if (this == o) {
      return true;
    }
    if (o == null || getClass() != o.getClass()) {
      return false;
    }
    AbstractVerifier that = (AbstractVerifier) o;
    return verificationType == that.verificationType;
  }

  @Override
  public int hashCode() {
    return Objects.hash(verificationType);
  }
}
